﻿using UnityEngine;
using System;
using System.Collections.Generic;
using static MacGruber.Breathing;
using static MacGruber.Utils;

namespace MacGruber
{	
	// Force override needed to let the Life plugin find its package
	public static partial class Utils
    {
		static Utils()
		{
			PackagePathOverride = "MacGruber.Life.12:/";  // <= Make sure to refer to the correct package here!
			PluginPathOverride = PackagePathOverride + "Custom/Scripts/MacGruber/Life/";
			IsInPackageOverride = true;
		}
	}

	public class DriverBreathingNoMouth : MVRScript
	{
		private Breathing myBreathing;

		private JSONStorableFloat myStomachMin;
		private JSONStorableFloat myStomachMax;
		private JSONStorableFloat myChestJointMin;
		private JSONStorableFloat myChestJointMax;
		private JSONStorableFloat myChestMorphMin;
		private JSONStorableFloat myChestMorphMax;
		private JSONStorableFloat myChestSpring;
		private JSONStorableFloat myNoseInMax;
		private JSONStorableFloat myNoseOutMax;
		private JSONStorableFloat myDebug1;
		private JSONStorableFloat myDebug2;
		private JSONStorableFloat myDebug3;

		DAZCharacterSelector myGeometry;
		DAZCharacterSelector.Gender myGender = DAZCharacterSelector.Gender.None;
		private DAZMorph myChestMorph;
		private DAZMorph myStomachMorph;
		private DAZMorph myNoseInMorph;
		private DAZMorph myNoseOutMorph;
		private FreeControllerV3 myChestControl;
		private bool myAnimationValid = false;
		private float myLungScale = 0.0f;
		private float myLungValuePrev = 0.0f;
		private bool myLungWasDown = true;
		private float myNoseScale = 0.0f;
		private float myNoseSign = 0.0f;

		private BlendValue myLungValue;
		private BlendValue myBreathInValue;
		private BlendValue myBreathOutValue;
		private BlendValue myDebugValue1;
		private BlendValue myDebugValue2;
		private BlendValue myDebugValue3;

		public override void Init()
		{	
			myBreathing = FindWithinSamePlugin<Breathing>(this);
			InitUI();
			InitAnimation();
			InitEvents();
		}

		private void InitUI()
		{
			Utils.OnInitUI(CreateUIElement);
			Utils.SetupInfoTextNoScroll(this,
				"<color=#606060><size=40><b>DriverBreathingNoMouth</b></size>\nThis is a custom breathing animation driver for Life that does not have mouth/lips animation.</color>\n\n" +
				"<b>ChestMorph Min/Max:</b>\nMinimum/Maximum values for chest morph.\n\n" +
				"<b>ChestJointDrive Min/Max:</b>\nMinimum/Maximum values for 'Joint Drive X Angle' of chest control.\n\n" +
				"<b>ChestJointDrive Spring:</b>\nValue for 'Joint Drive Spring' of chest control.\n\n" +
				"<b>Stomach Min/Max:</b>\nMinimum/Maximum values for stomach morph.\n\n" +
				"<b>NoseInMorph Max:</b>\nMaximum value for nose morph when breathing in.\n\n" +
				"<b>NoseOutMorph Max:</b>\nMaximum value for nose morph when breathing out.\n\n",
				1350.0f, true
			);

			myChestMorphMin = SetupSliderFloat(this, "ChestMorph Min", -0.1f, -1.0f, 1.0f, false);
			myChestMorphMax = SetupSliderFloat(this, "ChestMorph Max",  0.5f, -1.0f, 1.0f, false);
			myChestJointMin = SetupSliderFloat(this, "ChestJointDrive Min", -2.0f, -20.0f, 20.0f, false);
			myChestJointMax = SetupSliderFloat(this, "ChestJointDrive Max", 3.0f, -20.0f, 20.0f, false);
			myChestSpring = SetupSliderFloat(this, "ChestJointDrive Spring", 160.0f, 0.0f, 250.0f, false);
			myStomachMin = SetupSliderFloat(this, "StomachMin", -0.1f, -1.0f, 1.0f, false);
			myStomachMax = SetupSliderFloat(this, "StomachMax",  0.4f, -1.0f, 1.0f, false);
			myNoseInMax = SetupSliderFloat(this, "NoseInMorph Max", 0.6f, 0.0f, 1.0f, false);
			myNoseOutMax = SetupSliderFloat(this, "NoseOutMorph Max", 0.25f, 0.0f, 1.0f, false);

			if (!Utils.IsInPackage(this))
			{
				myDebug1 = SetupSliderFloat(this, "Debug1", 0.0f, 0.0f, 1.0f, true);
				myDebug2 = SetupSliderFloat(this, "Debug2", 0.0f, 0.0f, 1.0f, true);
				myDebug3 = SetupSliderFloat(this, "Debug3", 0.0f, 0.0f, 1.0f, true);
			}
        }

		private void InitAnimation()
		{
			myGeometry = containingAtom.GetStorableByID("geometry") as DAZCharacterSelector;
			string morphDirMale = "Custom/Atom/Person/Morphs/male/MacGruber/Pose/Breathing/";
			string morphDirFemale = "Custom/Atom/Person/Morphs/female/MacGruber/Pose/Breathing/";
			ResetMorphs(myGeometry?.morphsControlMaleUI, morphDirMale);
			ResetMorphs(myGeometry?.morphsControlFemaleUI, morphDirFemale);

			myGender = myGeometry != null ? myGeometry.gender : DAZCharacterSelector.Gender.None;
			GenerateDAZMorphsControlUI morphControl = myGeometry?.morphsControlUI;
			if (morphControl != null)
			{
				string morphDir = Utils.GetPackagePath(this) + (myGender == DAZCharacterSelector.Gender.Male ? morphDirMale : morphDirFemale);

				myChestMorph   = morphControl.GetMorphByUid(morphDir + "Breathing_Chest.vmi");
				myStomachMorph = morphControl.GetMorphByUid(morphDir + "Breathing_Stomach.vmi");
				myNoseInMorph  = morphControl.GetMorphByUid(morphDir + "Breathing_NoseIn.vmi");
				myNoseOutMorph = morphControl.GetMorphByUid(morphDir + "Breathing_NoseOut.vmi");
			}
			myChestControl = containingAtom.GetStorableByID("chestControl") as FreeControllerV3;

			myAnimationValid =
				myChestMorph     != null &&
				myStomachMorph   != null &&				
				myNoseInMorph    != null &&
				myNoseOutMorph   != null &&
				myChestControl   != null;
		}

		private void ResetMorphs(GenerateDAZMorphsControlUI morphControl, string morphDir)
		{
			if (morphControl == null)
				return;

			List<DAZMorph> morphs = morphControl.GetMorphs();
			for (int i=0; i<morphs.Count; ++i)
			{
				if (morphs[i].uid.Contains(morphDir))
					morphs[i].Reset();
			}
		}

		private void OnDisable()
		{
			if (myAnimationValid)
			{
				myChestControl.jointRotationDriveXTarget = 0.0f;
				myChestMorph.Reset();
				myStomachMorph.Reset();
				myNoseInMorph.Reset();
				myNoseOutMorph.Reset();
			}
		}

		private void OnDestroy()
        {
			Utils.OnDestroyUI();
        }

		private void InitEvents()
		{
			{
				BlendEvent value1 = myBreathing.RegisterBlendEvent(
					SM(MRK_BreathIn),
					SM(MRK_HoldIn),
					SM(MRK_BreathEnd)
				);
				BlendEvent value2 = myBreathing.RegisterBlendEvent(
					SM(MRK_BreathOut),
					SM(MRK_HoldOut),
					SM(MRK_BreathEnd)
				);

				myLungValue = new BlendCombineLinearInOut(value1, value2);

				myDebugValue1 = value1;
				myDebugValue2 = value2;
				myDebugValue3 = myLungValue;
			}

			{
				BlendEvent value1 = myBreathing.RegisterBlendEvent(
					SM(MRK_BreathIn, -0.12f),
					RM(MRK_BreathIn, MRK_HoldIn, 0.70f, 0.00f, 1.00f),
					SM(MRK_BreathEnd)
				);
				BlendEvent value2 = myBreathing.RegisterBlendEvent(
					RM(MRK_HoldIn, MRK_BreathIn, 0.30f, -0.40f, -0.05f),
					SM(MRK_HoldIn,  0.1f),
					SM(MRK_BreathEnd)
				);
				myBreathInValue = new BlendCombineLinearInOut(value1, value2);
			}

			{
				BlendEvent value1 = myBreathing.RegisterBlendEvent(
					SM(MRK_BreathOut, -0.12f),
					RM(MRK_BreathOut, MRK_HoldOut, 0.30f, 0.0f, 0.2f),
					SM(MRK_BreathEnd)
				);
				BlendEvent value2 = myBreathing.RegisterBlendEvent(
					RM(MRK_HoldOut, MRK_BreathOut, 0.70f, -2.0f, 0.0f),
					SM(MRK_HoldOut,  0.05f),
					SM(MRK_BreathEnd)
				);
				myBreathOutValue = new BlendCombineLinearInOut(value1, value2);
			}
		}

		private void Update()
		{
			if (myGeometry == null)
				return;
			if (myGeometry.gender != myGender)
				InitAnimation();

			if (myDebug1 != null && myDebugValue1 != null)
				myDebug1.val = myDebugValue1.Value;
			if (myDebug2 != null && myDebugValue2 != null)
				myDebug2.val = myDebugValue2.Value;
			if (myDebug3 != null && myDebugValue3 != null)
				myDebug3.val = myDebugValue3.Value;

			Breath breath = myBreathing.Get(CURR);
			if (!breath.IsValid() || !myAnimationValid)
				return;

			{
				float lungValue = myLungValue.Value;
				if (lungValue >= myLungValuePrev && myLungWasDown)
					myLungScale = breath.Depth * breath.Power * breath.Lung;
				myLungWasDown = lungValue < myLungValuePrev;
				myLungValuePrev = lungValue;
				lungValue = Mathf.SmoothStep(0.0f, myLungScale, lungValue);

				// correct a bit for the chest and stomach morphs being essentially linear scale, moving faster near max range
				float lungValueCorrected = Mathf.Pow(lungValue, 0.9f);

				// Chest morphs
				myChestMorph.morphValue = Mathf.Lerp(myChestMorphMin.val, myChestMorphMax.val, lungValueCorrected);

				// Chest joint rotation
				float chestJoint = Mathf.Lerp(myChestJointMin.val, myChestJointMax.val, lungValue);
                myChestControl.jointRotationDriveXTarget = chestJoint;
				myChestControl.jointRotationDriveDamper = 1.0f;
				myChestControl.jointRotationDriveSpring = myChestSpring.val;

				// Stomach morph
				myStomachMorph.morphValue = Mathf.Lerp(myStomachMin.val, myStomachMax.val, lungValueCorrected);
            }

			{
				float nose = 0.0f;
				bool hasNoseIn = breath.Entry.HasTag(TAG_NOSEIN);
				bool hasNoseOut = breath.Entry.HasTag(TAG_NOSEOUT);
				if (hasNoseIn && hasNoseOut)
				{
					nose = myBreathInValue.Value - myBreathOutValue.Value;
				}
				else if (hasNoseIn)
				{
					nose = myBreathInValue.Value;
				}
				else if (hasNoseOut)
				{
					nose = -myBreathOutValue.Value;
				}

				// Nose morphs
				float noseSign = Mathf.Sign(nose);
				if (myNoseSign != noseSign)
				{
					myNoseScale = breath.Depth * breath.Power * breath.Nose;
					myNoseSign = noseSign;
				}
				if (nose >= 0.0f)
				{
					myNoseInMorph.morphValue = myNoseScale*myNoseInMax.val*nose;
					myNoseOutMorph.morphValue = 0.0f;
				}
				else
				{
					myNoseInMorph.morphValue = 0.0f;
					myNoseOutMorph.morphValue = -myNoseScale*myNoseOutMax.val*nose;
				}
			}
		}
	}
}